/*
  2 threads. Pagecache has N blocks.
  thd1 calls pagecache_write() for page 0, ... page N-1.
  At this point there are no more free blocks.
  thd2 calls pagecache_read() for page N,
  this calls find_block(): as there are no free blocks, it evicts the
  least-recently-used one which is page 0: it writes page 0 to disk, after the
  pwrite() we make it pause.
  thd1 calls pagecache_read for page 0, removes the IN_SWITCH (which looks
  like a wrong thing to do).
  Then thd1 calls flush_pagecache(FLUSH_IGNORE_CHANGED), and crashes (page 0
  is not in LRU).
*/

#include <tap.h>
#include <my_sys.h>
#include <m_string.h>
#include "../ma_pagecache.h"

// small page cache for fast test and small debug trace, but
// fails the same way with big page cache.
#define PCACHE_SIZE (TEST_PAGE_SIZE*16)

#ifndef DBUG_OFF
static const char* default_dbug_option;
#endif

static char *file1_name= (char*)"page_cache_test_file_1";
static PAGECACHE_FILE file1;
static pthread_cond_t COND_thread_count;
static pthread_mutex_t LOCK_thread_count;
static uint thread_count;
static PAGECACHE pagecache;

int pagen;

/**
  @brief Dummy pagecache callback.
*/

static my_bool
dummy_callback(uchar *page __attribute__((unused)),
               pgcache_page_no_t page_no __attribute__((unused)),
               uchar* data_ptr __attribute__((unused)))
{
  return 0;
}


/**
  @brief Dummy pagecache callback.
*/

static void
dummy_fail_callback(uchar* data_ptr __attribute__((unused)))
{
  return;
}

/* variables to signal state changes between threads */
int step2=0, step3=0;

void simple_write_test()
{
  unsigned char *buff= malloc(TEST_PAGE_SIZE), *resbuff;
  int i, res;
  DBUG_ENTER("simple_write_test");

  // fill page cache with dirty pages
  for (i= 0; i<pagen; i++)
  {
    res= pagecache_write(&pagecache, &file1, i, 0, buff,
                         PAGECACHE_PLAIN_PAGE,
                         PAGECACHE_LOCK_LEFT_UNLOCKED,
                         PAGECACHE_PIN_LEFT_UNPINNED,
                         PAGECACHE_WRITE_DELAY,
                         0, LSN_IMPOSSIBLE);
    assert(!res);
  }

  step2=1;

  while(!step3)
    sleep(1);

  sleep(3); // time for thread2 to block during flush

  // now let's flush
  res= pagecache_delete(&pagecache, &file1, 0, PAGECACHE_LOCK_WRITE, 0);
  // we should have crashed here already, if bug exists.

  assert(!res);

  free(buff);
  DBUG_VOID_RETURN;
}

void simple_read_test()
{
  unsigned char *buff= malloc(TEST_PAGE_SIZE);
  char *resbuff;
  DBUG_ENTER("simple_read_test");

  // wait for pagecache to be full of dirty pages
  while (!step2)
    sleep(1);

  // signal that we start flush of page
  step3=1;
#if 1 // use eviction
  // evict one
  resbuff= pagecache_read(&pagecache, &file1, pagen, 0, buff,
                          PAGECACHE_PLAIN_PAGE,
                          PAGECACHE_LOCK_LEFT_UNLOCKED,
                          0);
#else // use full flush
  flush_pagecache_blocks(&pagecache, &file1, FLUSH_KEEP);
#endif
  // should never come here, as sleep is long.
  assert(0);

  free(buff);
  DBUG_VOID_RETURN;
}

static void *thread1_func(void *arg)
{
  my_thread_init();

  simple_write_test();

  DBUG_PRINT("info", ("Thread %s ended\n", my_thread_name()));
  pthread_mutex_lock(&LOCK_thread_count);
  thread_count--;
  VOID(pthread_cond_signal(&COND_thread_count)); /* Tell main we are ready */
  pthread_mutex_unlock(&LOCK_thread_count);
  free((uchar*) arg);
  my_thread_end();
  return NULL;
}

static void *thread2_func(void *arg)
{
  my_thread_init();

  simple_read_test();

  DBUG_PRINT("info", ("Thread %s ended\n", my_thread_name()));
  pthread_mutex_lock(&LOCK_thread_count);
  thread_count--;
  VOID(pthread_cond_signal(&COND_thread_count)); /* Tell main we are ready */
  pthread_mutex_unlock(&LOCK_thread_count);
  free((uchar*) arg);
  my_thread_end();
  return NULL;
}


int main(int argc, char **argv)
{
  pthread_t tid;
  pthread_attr_t thr_attr;
  int error;
  MY_INIT(argv[0]);

#ifndef DBUG_OFF
#if defined(__WIN__)
  default_dbug_option= "d:t:i:O,\\test_pagecache_single.trace";
#else
  default_dbug_option= "d:t:i:O,/tmp/test_pagecache_single.trace";
#endif
  if (argc > 1)
  {
    DBUG_SET(default_dbug_option);
    DBUG_SET_INITIAL(default_dbug_option);
  }
#endif

  DBUG_ENTER("main");
  DBUG_PRINT("info", ("Main thread: %s\n", my_thread_name()));

  if ((file1.file= my_open(file1_name,
                           O_CREAT | O_TRUNC | O_RDWR, MYF(0))) == -1)
  {
    fprintf(stderr, "Got error during file1 creation from open() (errno: %d)\n",
	    errno);
    exit(1);
  }
  pagecache_file_init(file1, &dummy_callback, &dummy_callback,
                      &dummy_fail_callback, &dummy_callback, NULL);

  DBUG_PRINT("info", ("file1: %d", file1.file));
  if (my_chmod(file1_name, S_IRWXU | S_IRWXG | S_IRWXO, MYF(MY_WME)))
    exit(1);

  if ((error= pthread_cond_init(&COND_thread_count, NULL)))
  {
    fprintf(stderr, "Got error: %d from pthread_cond_init (errno: %d)\n",
	    error, errno);
    exit(1);
  }
  if ((error= pthread_mutex_init(&LOCK_thread_count, MY_MUTEX_INIT_FAST)))
  {
    fprintf(stderr, "Got error: %d from pthread_cond_init (errno: %d)\n",
	    error, errno);
    exit(1);
  }

  if ((error= pthread_attr_init(&thr_attr)))
  {
    fprintf(stderr,"Got error: %d from pthread_attr_init (errno: %d)\n",
	    error,errno);
    exit(1);
  }
  if ((error= pthread_attr_setdetachstate(&thr_attr, PTHREAD_CREATE_DETACHED)))
  {
    fprintf(stderr,
	    "Got error: %d from pthread_attr_setdetachstate (errno: %d)\n",
	    error,errno);
    exit(1);
  }

#ifdef HAVE_THR_SETCONCURRENCY
  VOID(thr_setconcurrency(2));
#endif

  if ((pagen= init_pagecache(&pagecache, PCACHE_SIZE, 0, 0,
                             TEST_PAGE_SIZE, MYF(MY_WME))) == 0)
  {
    fprintf(stderr,"Got error: init_pagecache() (errno: %d)\n",
            errno);
    exit(1);
  }
  DBUG_PRINT("info", ("Page cache %d pages", pagen));

  pthread_mutex_lock(&LOCK_thread_count);
  if ((error= pthread_create(&tid, &thr_attr, thread1_func, NULL)))
  {
    fprintf(stderr,"Got error: %d from pthread_create (errno: %d)\n",
            error,errno);
    exit(1);
  }
  if ((error= pthread_create(&tid, &thr_attr, thread2_func, NULL)))
  {
    fprintf(stderr,"Got error: %d from pthread_create (errno: %d)\n",
            error,errno);
    exit(1);
  }
  thread_count= 2;
  DBUG_PRINT("info", ("Threads started"));
  pthread_mutex_unlock(&LOCK_thread_count);

  pthread_attr_destroy(&thr_attr);

  pthread_mutex_lock(&LOCK_thread_count);
  while (thread_count)
  {
    if ((error= pthread_cond_wait(&COND_thread_count,&LOCK_thread_count)))
      fprintf(stderr,"Got error: %d from pthread_cond_wait\n",error);
  }
  pthread_mutex_unlock(&LOCK_thread_count);
  DBUG_PRINT("info", ("thread ended"));

  end_pagecache(&pagecache, 1);
  DBUG_PRINT("info", ("Page cache ended"));

  if (my_close(file1.file, MYF(MY_WME)))
    exit(1);

  DBUG_PRINT("info", ("Program end"));

  my_end(0);

  return exit_status();
}
